/*
 * Gateway to implementation of ProjMP2D in c++
 *
 * SPMP2D Self Projected Matching Pursuit 2D
 * 
 * It projects MP2D approximation at every "lstep" iterations using 
 * Projected  Matching  Pursuit 2D (ProjMP2D)
 * For lstep=1 it converges to Orthogonal Matching Pursuit (OMP) 
 * For lstep=0 it projects MP2D only once to improve upon the final MP2D 
 * approximation
 * For lstep=-1 it gives plain MP with no projection at all 
 *
 * Usage: [h, Di1, Di2,c] = SPMP2D(f,Vx,Vy,tol,No,toln,lstep,Max,...
 *              Maxp,indx,indy);
 *        [h, D1, D2,c] = SPMP2D(f,Vx,Vy);
 *
 * Inputs:
 *  f     2D signal (Image)
 *  Vx    dictionary of normalized 1D atoms (image's raws)
 *  Vy    dictionary of normalized 1D atoms (image's columns)
 *  tol   tolerance for the approximation (ems = (error norm)^2/numel(f))  
 *  No    maximum number of atoms in the approximation
 *  toln  numerical tolerance for the projection  (default tre=1e-3)
 *  lstep length of the steps for projectting (default lstep=0)
 *  Max   maximum number of iterations  (default Max=7000)
 *  Maxm  maximum number of iterations for projection (default Max=7000)
 *  indx  (optional) indices for an initial subspace; They operate as indx(k),indy(k)
 *  indy  (optional) indices for an initial subspace
 *
 * Outputs:
 *  h     approximation of f (Image)
 *  Di1   indexes of selected (distinct) atoms (w.r.t. the original Vx)
 *  Di2   indexes of selected (distinct) atoms (w.r.t. the original Vy)
 *  c     coefficeints of the atomic decomposition: sum_n(c(n)*Vx(:,Di1(n))*Vy(:,Di2(n))
 */

#define output 0
typedef unsigned long int ulng;
#include "mex.h"
#include "SPMP2D.cpp"
using namespace std;

/* Function Declarations */
static void SPMP2D_Mex( int nlhs, mxArray *plhs[],
        int nrhs, const mxArray *prhs[]);
double *GetImage(const mxArray *prhs,unsigned long &nRows,
        unsigned long &nColumns);
double *GetVIndDictColumn(const mxArray *prhs, ulng nAtoms, ulng &nInd);
ulng ValidateVIndDictC(unsigned long nAtoms, unsigned long nRows,
        unsigned long nColumns);
double *GetVIndDictRow(const mxArray *prhs, ulng nIndColumn);
void ValidateVIndDictR(unsigned long nIndColumn, unsigned long nRows,
        unsigned long nColumns);
double *GetDict(const mxArray *prhs, unsigned long nRowsToCheck,
        unsigned long &nAtoms);
void ValidateDict(unsigned long nRowsToCheck, unsigned long nRows);
double GetDouble(const mxArray *prhs);
ulng Getulng(const mxArray *prhs);
void ValidateScalar(unsigned long nRows, unsigned long nColumns);
double *GetVar(const mxArray *prhs, unsigned long &M,
        unsigned long &N, const char *varName);
double *SetVar(mxArray *&m, unsigned long nRows, unsigned long 
        nColumns);
ulng ValidateVector(ulng n, ulng m);
long GetLong(const mxArray *prhs);

void mexFunction(int nlhs, mxArray *plhs[],
    int nrhs, const mxArray *prhs[])
{
    /* Gateway from MATLAB */
    SPMP2D_Mex(nlhs,plhs,nrhs,prhs);
}

static void SPMP2D_Mex( int nlhs, mxArray *plhs[],
        int nrhs, const mxArray *prhs[])
{
    /* Validate input to function */
        
    /* Validate number of input (rhs) arguments*/    
    if((nrhs < 3) || (nrhs > 11))
    {
        mexErrMsgTxt("Wrong number of input arguments");
    }
    
    /* Check we are being passed real full and double variables */
    for ( int i = 0; i < nrhs; i++ )
    {        
        if ( mxIsComplex(prhs[i]) || mxIsClass(prhs[i],"sparse")|| 
                !mxIsDouble(prhs[i]) )
        {
            mexErrMsgTxt("Input must be double, real, full");
        }
    }
    
        
    /*
     * Get the manditory inputs.
     */
        
    /* Check enough (rhs) arguments*/    
 //   if(nrhs < 4 )
  //  {
 //       mexErrMsgTxt("Wrong number of input arguments");
 //   }
    
    /* Get the 2D image
     */
    unsigned long nRows = 0;
    unsigned long nColumns = 0;
    double *mImage = GetImage(prhs[0],nRows,nColumns);

    
   /* Get 1D Dictionary (horizontal coordinate) performs 
     inner products with the columns of the residule 
    */
   unsigned long nAtomsColumn = 0;
   double *mDictColumns = GetDict(prhs[1],nRows,nAtomsColumn);
    
    /* Get 1D Dictionary (vertical coordinate) performs 
     inner products with the rows of the residule 
     */
    ulng nAtomsRow = 0;
    double *mDictRows = GetDict(prhs[2],nColumns,nAtomsRow);     
    
    
    /* Number of pixels
     *  Calculated here assuming validation is passes
     */
    ulng nPixels = nRows*nColumns;
    
    
    /* 
     * Check for the optional inputs
     */
    // <<---
        /* Get the numerical tolerance for approximation
         * MATLAB
         *  if (nargin<4)  | (isempty(tol)==1)   
         *      tol=5e-4*sum(sum(abs(f).^2));
         *  end
         */
        double tolApprox = 0;
//         if (nrhs > 3)
//         {
//             if (mxIsEmpty(prhs[3]))
//             {
//                 tolApprox = 5e-4* Norm(mImage,nPixels);
//             }
        // If their is a variable and its dimensions are not zero retreive it
        if ((nrhs > 3) && !mxIsEmpty(prhs[3]))
        {
            tolApprox = GetDouble(prhs[3]);
        }else{
            /* tol=5e-4*sum(sum(abs(f).^2)); 
             * CHANGED TO
             * tol = 5e-4*norm(f,'fro');
             */
            tolApprox = 5e-4* Norm(mImage,nPixels);
        }
    
        
        /* Get maximum number of atoms in the approximation
         * MATLAB
         *  if (nargin<5)  | (isempty(No)==1)   
         *      No=Nxy;
         *  end
         */
        ulng maxNAtoms = 0;
        if ((nrhs > 4) && !mxIsEmpty(prhs[4]))
        {
            maxNAtoms = Getulng(prhs[4]);
        }else{
            maxNAtoms = nPixels;
        }        
        
        /* Get numerical tolerance for the projection  (default tre=1e-3)
         *  if (nargin<6)  | (isempty(toln)==1) 
         *      tolnu=1e-3;
         *  end
         */
        double tolProjection = 0;
        if ((nrhs > 5) && !mxIsEmpty(prhs[5]))
        {
            tolProjection = GetDouble(prhs[5]);
        }else{
            tolProjection = 1e-3;
        }        
                
        /* Get length of the steps for projectting (default lstep=0)
         * MATLAB CODE
         *  if (nargin<7)  | (isempty(lstep)==1) 
         *      lstep=0;
         *  end
         */
        long lStep = 0;
        if ((nrhs > 6) && !mxIsEmpty(prhs[6]))
        {
            lStep = GetLong(prhs[6]);
        }  
        
        
        /* Get maximum number of iterations  (default Max=7000)
         * MATLAB
         *  if (nargin<8)  | (isempty(Max)==1)   
         *      Max=7000;
         *  end
         */
        ulng maxIter = 7000;
        if ((nrhs > 7) && !mxIsEmpty(prhs[7]))
        {
            maxIter = Getulng(prhs[7]);
        }     
        
        
        /* Get maximum number of iterations for projection (default Max=7000)
         * MATLAB
         *  if (nargin<9) | (isempty(Maxp)==1)  
         *      Maxp=7000;
         *  end
         */
        ulng maxIterProj = 7000;
        if ((nrhs > 8) && !mxIsEmpty(prhs[8]))
        {
            maxIterProj = Getulng(prhs[8]);
        }         
                

		/* Indexes for both initial subspaces should be provided */
		if (nrhs == 10)
		{
			mexErrMsgTxt("Initial indexes should be provided for both dictionaries!\n");
		}


       /* Get indx (optional) indices for an initial subspace;
        *  MATLAB
        *   if (nargin<10) | (isempty(indx)==1)  
        *       indx=[];
        *   end  
        *   if (nargin<11) | (isempty(indy)==1)  
        *       indy=[];
        *   end  
		*/

        double *vIndDictColumns = 0;
		double *vIndDictRows = 0;
        ulng nInitialIndexes = 0;
        if (nrhs > 10)
        {
            vIndDictColumns = GetVIndDictColumn(prhs[9],nAtomsColumn,
                    nInitialIndexes);

			vIndDictRows = GetVIndDictRow(prhs[10],nInitialIndexes);

        } 

    // --->>
        
        
        
        
    /* Validate number of output (lhs) arguments*/    
    if ((nlhs<4) || (nlhs>5))
    {
        mexErrMsgTxt("Wrong number of output arguments");
    }
    
    /* Initialize Output variables */
    // <<--
        /* Initialize arrays for output data */

        /* Create mxArray for the approximation */
        double *mApprox = SetVar(plhs[0],nRows,nColumns);

        /* Create a dynamic array large enough to store the chosen column 
         * dictionary indexes - will not exceed nPixels
         */
        double *vIColumnAtomsBig = new double[nPixels];
        
        /* Create a dynamic array large enough to store the chosen row 
         * dictionary indexes
         */
        double *vIRowAtomsBig = new double[nPixels];        
        
        /* Create a dynamic array large enough to store the chosen 
         * coefficients in the approximation
         */
		// Use nPixels as maxNAtoms may be too large 
        double *vCoefficientsBig = new double[nPixels];    
    // -->>
    
  
    /* # of Atoms chosen in target function */
    ulng nChosen = 0;
    
    /* # of projections performed by ProjMP2D */
    ulng projMPIter = 0;
    
    /* Invoke target function */
    SPMP2D(mImage,mDictColumns,mDictRows,tolApprox,maxNAtoms,tolProjection,
            lStep, maxIter,maxIterProj,vIndDictColumns,vIndDictRows,
            mApprox,vIColumnAtomsBig,vIRowAtomsBig,vCoefficientsBig
            ,nChosen,nInitialIndexes,
            nRows,nColumns,nAtomsColumn,nAtomsRow,projMPIter);

    
    /* Initialize and populate remaining mxArrays now we have nChosen */
    // <<--
        double *vIColumnAtoms = SetVar(plhs[1],1,nChosen);
        
        /* Copy accros chosen indexes */
        CopyElements(vIColumnAtomsBig,vIColumnAtoms,nChosen);  
        
        double *vIRowAtoms = SetVar(plhs[2],1,nChosen);
        
        /* Copy accros chosen indexes */
        CopyElements(vIRowAtomsBig,vIRowAtoms,nChosen); 
        
        double *vCoefficients = SetVar(plhs[3],1,nChosen);
        
        /* Copy accros chosen indexes */
        CopyElements(vCoefficientsBig,vCoefficients,nChosen);
    // -->>
        
    /* Optional output params */
        if (nlhs == 5)
        {
            /* Initialize optional iteration counter */
            double *projMPIterOut = SetVar(plhs[4],1,1);
            *projMPIterOut = projMPIter;
        }      
        
    /* Free dynamicly assigned memory */
    // <<--
        delete [] vIRowAtomsBig;
        vIRowAtomsBig = 0;
        delete [] vIColumnAtomsBig;
        vIRowAtomsBig = 0;        
        delete [] vCoefficientsBig;
        vCoefficientsBig = 0;  
    // -->>
}

double *GetImage(const mxArray *prhs,unsigned long &nRows,
        unsigned long &nColumns)
{
    /* Get the image */
    double *mImage = GetVar(prhs,nRows,nColumns,"Image");
    
    return mImage;
    
}

double *GetVIndDictColumn(const mxArray *prhs, ulng nAtoms, ulng &nInd)
{
	double *vIndDict  = 0;
	if (mxIsEmpty(prhs))
	{
		nInd = 0;
		vIndDict = 0;
	}else{

		/* Get the initial indexes chosen from the column dictionary*/
		unsigned long nRows = 0;
		unsigned long nColumns = 0;
		vIndDict = GetVar(prhs,nRows,nColumns,"IndDictColumn");  
    
		nInd = ValidateVIndDictC(nAtoms,nRows,nColumns);
	}
    
    return vIndDict;
    
}

ulng ValidateVIndDictC(unsigned long nAtoms, unsigned long nRows,
        unsigned long nColumns)
{
    /* Checks that the number of indexes is less than or equal to the 
     * number of atoms */
    

    /* The largest dimension is the number of indexes */
    ulng nIndexes = ValidateVector(nRows, nColumns);
        
    
    /* Validate */
    if (nIndexes > nAtoms)
    {
        mexPrintf("# of atoms %i, # of indexes %i\n",
            nAtoms, nIndexes);
        mexErrMsgTxt("Number of indexes should be less than the number of atoms!!");                
    }
    
    return nIndexes;
}



double *GetVIndDictRow(const mxArray *prhs, ulng nIndColumn)
{
	double *vIndDict = 0;
	if ((mxIsEmpty(prhs)) && (nIndColumn == 0))
	{
		vIndDict = 0;
	}else{

		/* Get the initial indexes chosen from the column dictionary*/
		unsigned long nRows = 0;
		unsigned long nColumns = 0;
		vIndDict = GetVar(prhs,nRows,nColumns,"IndDictRow");  
    
		ValidateVIndDictR(nIndColumn,nRows,nColumns);
	}
    
    return vIndDict;
    
}

void ValidateVIndDictR(unsigned long nIndColumn, unsigned long nRows,
        unsigned long nColumns)
{
    /* Checks that the number of indexes is less than or equal to the 
     * number of atoms */
    
    /* The largest dimension is the number of indexes */
    ulng nIndexes = ValidateVector(nRows, nColumns);
        
    
    /* Validate */
    if (nIndexes != nIndColumn)
    {
        mexPrintf("# of initial column indexes %i, # of intitial row indexes %i\n",
            nIndColumn, nIndexes);
        mexErrMsgTxt("Number of indexes should be the same!!");                
    }

}


ulng ValidateVector(ulng n, ulng m)
{
    /* Checks for singleton dimension and returns number of elements in the
     * vector.
     */
    ulng nElements = 0;
    if ((n == 1) && (m >= 1))
    {
        nElements = m;
    }else if( (m==1) && (n >= 1))
        nElements = n;
    else
    {
        mexPrintf("Vector dimensions %ix%i\n",n,m);
        mexErrMsgTxt("Not a vector!!!");
    }
    
    return nElements;
}

double *GetDict(const mxArray *prhs, unsigned long nRowsToCheck,
        unsigned long &nAtoms)
{
    /* Get the dictionary */
    unsigned long nRows = 0;
    //unsigned long nColumn = 0;
    double *mDict = GetVar(prhs,nRows,nAtoms,
            "Dictionary");    
    
    /* Validate dimensions */
    ValidateDict(nRowsToCheck,nRows);
    
    return mDict;
    
}

void ValidateDict(unsigned long nRowsToCheck, unsigned long nRows)
{
    if (nRowsToCheck != nRows)
    {
        mexPrintf("Dictionary has %i rows, validate against %i rows\n",
                nRowsToCheck,nRows);
        mexErrMsgTxt("Error # rows in dictionary not consistent!!!!");
    }
    
}

double GetDouble(const mxArray *prhs)
{
    /* Get double scalar*/
    unsigned long nRows = 0;
    unsigned long nColumns = 0;

    double scalar = *GetVar(prhs,nRows,nColumns,"Scalar");

    /* Validate scalar */
    ValidateScalar(nRows,nColumns);
    
    return scalar;
    
}

ulng Getulng(const mxArray *prhs)
{
    /* Get double scalar*/
    unsigned long nRows = 0;
    unsigned long nColumns = 0;

    ulng scalar = static_cast<ulng>(*GetVar(prhs,nRows,nColumns,"Long"));

    /* Validate scalar */
    ValidateScalar(nRows,nColumns);
    
    return scalar;
    
}

long GetLong(const mxArray *prhs)
{
    /* Get double scalar*/
    unsigned long nRows = 0;
    unsigned long nColumns = 0;

    long scalar = static_cast<long>(*GetVar(prhs,nRows,nColumns,"Long"));

    /* Validate scalar */
    ValidateScalar(nRows,nColumns);
    
    return scalar;
    
}

void ValidateScalar(unsigned long nRows, unsigned long nColumns)
{
    
    if ((nRows != 1) || (nColumns != 1))
    {
        mexPrintf("Variable has %i rows and %i columns\n",nRows, 
                nColumns);
        mexErrMsgTxt("Variable should be a scalar!!");
    }
    
}

double *GetVar(const mxArray *prhs, unsigned long &M,
        unsigned long &N, const char *varName)
{
    /* Gets a pointer to the double array stored in the mxArray prhs[0] 
     * and retrieves the dimensions of the array.
     *  Validates that dimensions are not empty (0).
     */
    
    /* Get pointer to array */
    double *var = mxGetPr(prhs);
    
    /* Get array dimensions */
    M = mxGetM(prhs);
    N = mxGetN(prhs);
    
    /* Validate that array is not empty */
     if ((M == 0) || (N == 0))
     {
         mexPrintf("Input array %s has zero dimensions\n",varName);
         mexErrMsgTxt("Error!!!!");
     }
     
    return var;

}


double *SetVar(mxArray *&m, unsigned long nRows, unsigned long 
        nColumns) // Use reference to pointer
{
    /* Creates a real mxArray, size nRows x nColumns assigns it to m[0]
     * and returns the 
     * pointer to it */
    
   m = mxCreateDoubleMatrix(nRows, nColumns, mxREAL);
    
  //  double *t = mxGetPr(m);
   // t[0] = 1;
    return mxGetPr(m);
}

